/*
 * Copyright 1994 Sun Microsystems, Inc. All Rights Reserved
 *
 * Use cc -L /opt/SUNWconn/x25/lib -o clnt clnt.c -lsx25 to compile this
 * program.
 *
 * Usage: clnt [-u <call user data>] <address> [<file_to_receive>]
 * This program is the calling side of an SVC.  The program sends a
 * Call Request to the remote host at <address> with the given user
 * data (hex 0x12 0x34 by default, otherwise the value given with the
 * '-u' flag).  After a connection is established with the remote server,
 * the program waits for normal data sent by the server side, and either
 * displays the received data on the screen, or writes it into the file
 * <file_to_receive> if this argument was given on the command line.
 * 
 * Example:  To connect over link 2, which is configured as an LLC2 link,
 * to a remote host with MAC address 08:00:20:0a:39:91, the command line
 * would be:
 *	clnt 2.0800200a39917e
 *
 * To test in loopback mode over LLC2, use the SAP '70' instead of the
 * normal SAP '7e'.  So to connect to the local host 08:00:20:0a:39:91, the
 * command line would be:
 *	clnt 2.0800200a399170
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <netx25/uint.h>
#include <netx25/timer.h>
#include <netdlc/ll_proto.h>
#include <netdlc/ll_control.h>
#include <netx25/x25_proto.h>
#include <netx25/x25_control.h>
#include <unistd.h>
#include <sys/time.h>

#define bzero(c,l)	memset(c,0,l)
#define bcopy(a,b,c)	memcpy(b,a,c)

/*
 * Maximum length of call user data (CUD) allowed to be entered by user.
 * (# of bytes)
 */
#define MAX_CUD_LEN	8


int
open_stream()
{

	int             x25_stream;

	x25_stream = open("/dev/x25", O_RDWR);

	if (x25_stream < 0) {
		perror("	***open");
		exit(1);
	}
	return (x25_stream);
}


/*
 * Send an N_CI message down to the X.25 driver.  This will cause a
 * Call Request packet to be sent out.
 */
void
call_request(s, address, cud, cudlen)
	int             s;
	struct xaddrf  *address;
	char           *cud;
	int             cudlen;
{
	int             flags = 0;
	struct strbuf   databuf;
	struct strbuf   ctlbuf;
	char            ctlblk[1024];
	char            datablk[1024];
	int             retval;

	struct xcallf  *call;
	struct xccnff  *caa;

	call = (struct xcallf *) malloc(sizeof(struct xcallf));
	bzero((char *) call, sizeof(struct xcallf));

	bzero((char *) ctlblk, 1024);
	bzero((char *) datablk, 1024);

	call->xl_command = N_CI;
	call->xl_type = XL_CTL;

	call->CONS_call = 0;

	if (address)
		bcopy((char *) address, (char *) &call->calledaddr, sizeof(struct xaddrf));
	else
		bzero((char *) &call->calledaddr, sizeof(struct xaddrf));

	ctlbuf.len = sizeof(struct xcallf);
	ctlbuf.buf = (char *) call;

	databuf.len = cudlen;
	databuf.buf = (char *) cud;


	/*
	 * Send a Call Request packet.
	 */

	if (putmsg(s, &ctlbuf, &databuf, 0) < 0) {
		perror("	***putmsg");
		exit(1);
	}

	ctlbuf.len = 0;
	ctlbuf.maxlen = 1024;
	ctlbuf.buf = (char *) ctlblk;

	databuf.len = 0;
	databuf.buf = datablk;
	databuf.maxlen = 1024;

	/*
	 * Wait for a response.  Normally we expect a Call Accept packet.
	 */
	retval = getmsg(s, &ctlbuf, &databuf, &flags);

	if (retval < 0) {
		perror("	***getmsg");
		exit(2);
	}
	caa = (struct xccnff *) ctlbuf.buf;

	if ((caa->xl_type != XL_CTL) || (caa->xl_command != N_CC)) {
		fprintf(stderr, "Connection refused\n");
		exit(3);
	}
	printf("Client received a Call Accept packet \n");
	printf("Successful connection to remote server \n");
	printf("\n");
}


/*
 * Initiate the connection to 'remote_peer', over the VC with file
 * descriptor 'virtual_circuit'.
 */
void
start_connection(remote_peer, virtual_circuit, cud, cud_len)
	char           *remote_peer;
	int             virtual_circuit;
	char		*cud;
	int		cud_len;
{
	struct xaddrf	called;

	bzero((char *) &called, sizeof(struct xaddrf));

	if (stox25(remote_peer, &called, 0)) {
		fprintf(stderr, "'%s': Bad address format\n", remote_peer);
		exit(1);
	}

	call_request(virtual_circuit, &called, cud, cud_len);
}


/*
 * Wait for and receive an N_Data message (Data packet) from the remote end.
 */
void
recv_normal_data(x25_stream, buffer, reqbytes, more, recvp)
	int             x25_stream;
	char           *buffer;
	int             reqbytes;
	int            *more;
	int            *recvp;
{
	struct strbuf   ct;
	struct strbuf   dt;
	int             flag = 0;
	int             retval;
	char            cblk[1024];
	S_X25_HDR      *result;

	bzero((char *) cblk, 1024);

	ct.buf = (char *) cblk;
	ct.maxlen = 1024;

	dt.buf = (char *) buffer;
	dt.maxlen = reqbytes;

	retval = getmsg(x25_stream, &ct, &dt, &flag);

	if (retval < 0) {
		perror("	***getmsg");
		exit(1);
	}
	result = (S_X25_HDR *) cblk;

	if ((result->xl_type != XL_DAT) || (result->xl_command != N_Data)) {

		if (result->xl_type == XL_CTL && result->xl_command == N_DI)
			fprintf(stderr, "Connection closed: End of data transfer\n");

		exit(2);
	}
	if (retval == MOREDATA)
		*more = MOREDATA;
	else
		*more = 0;

	*recvp = dt.len;
}


char		eof_string[] = "#!@# end of file";


/*
 * Keep receiving data from the remote end, and writing it to the file
 * 'fd', until the remote end indicates that it has finished (by sending
 * 'eof_string' as data).  Requiring the remote end to send the
 * 'eof_string' ensures that all the data is transferred, because
 * the N_DI (Clear Request) is destructive and would discard any data
 * queued in the STREAMS system that has not yet been sent out.
 */
void
data_transfer(vc, fd)
	int             vc;
	int             fd;
{
	int             m;
	int		cc;
	int             nbytes = 0;
	static char     buffer[1024];

	while (1) {
		recv_normal_data(vc, buffer, 1024, &m, &cc);
		if (cc == strlen(eof_string) &&
		    strncmp(buffer, eof_string, cc) == 0) {
			break;
		}
		write(fd, buffer, cc);
		nbytes += cc;
	}
	printf("\n");
	printf("Transfer:  %d bytes received\n", nbytes);
}


/*
 * Send an N_DI message down to the X.25 driver.  This will cause a
 * Clear Request packet to be sent out.
 */
void
send_clear_req(s)
	int		s;
{
	int             flags = 0;
	struct strbuf   databuf;
	struct strbuf   ctlbuf;
	char		datablk[1024];
	char            ctlblk[1024];
	int             retval;
	int		fpos;
	struct xdiscf	*disc;
	struct xdcnff	*dsc_conf;

	disc = (struct xdiscf *) malloc(sizeof(struct xdiscf));
	bzero((char *) disc, sizeof(struct xdiscf));

	bzero((char *) ctlblk, 1024);
	bzero((char *) datablk, 1024);

	disc->xl_command = N_DI;
	disc->xl_type = XL_CTL;
	disc->cause = 7;
	disc->diag = 42;

	disc->indicated_qos = 1;

	fpos = 0;

	disc->qos.xtras.fac_field[fpos++] = 0xc3;
	disc->qos.xtras.fac_field[fpos++] = 0x04;
	disc->qos.xtras.fac_field[fpos++] = 0x0f;
	disc->qos.xtras.fac_field[fpos++] = 0x04;
	disc->qos.xtras.fac_field[fpos++] = 0x12;
	disc->qos.xtras.fac_field[fpos++] = 0x34;

	disc->qos.xtras.x_fac_len = fpos;
	disc->qos.xtras.call_deflect = 1;


	/* All other fields of N_DI are left = 0. */

	ctlbuf.len = sizeof(struct xdiscf);
	ctlbuf.buf = (char *) disc;

	/*
	 * Send a Clear Request packet.
	 */
	if (putmsg(s, &ctlbuf, 0, 0) < 0) {
		perror("	***putmsg");
		exit(1);
	}

	ctlbuf.len = 0;
	ctlbuf.maxlen = 1024;
	ctlbuf.buf = (char *) ctlblk;

	databuf.len = 0;
	databuf.buf = datablk;
	databuf.maxlen = 1024;

	/*
	 * Wait for a response.  Normally we expect a Clear Confirm packet.
	 */
	retval = getmsg(s, &ctlbuf, &databuf, &flags);

	if (retval < 0) {
		perror("	***getmsg");
		exit(2);
	}
	dsc_conf = (struct xdcnff *) ctlbuf.buf;

	if ((dsc_conf->xl_type != XL_CTL) || (dsc_conf->xl_command != N_DC)) {
		fprintf(stderr, "Clear not confirmed\n");
		exit(3);
	}
	printf("Connection cleared\n");
}


/*
 * pack() takes as input a null-terminated ASCII string str representing
 * Call User data, and converts it to packed hex form.  'hexstr' points to
 * the array containing the packed output.  If the address has invalid
 * characters, -1 is returned.  Otherwise, the length of the address
 * (in hex digits) is returned.
 */
int
pack(str, hexstr)
	char           *str;
	u_char         *hexstr;
{
	int             i, j;
	u_char          c;

	i = j = 0;
	while (str[i]) {
		c = str[i];
		if (i >= MAX_CUD_LEN*2) {
			fprintf(stderr, "User data too long\n");
			return (-1);
		}
		if (!isxdigit(c)) {
			fprintf(stderr, "Invalid hex character '%c'\n", c);
			return (-1);
		}
		if (isdigit(c)) {
			c = c - '0';
		} else {
			if (isupper(c)) {
				c = tolower(c);
			}
			c = (c - 'a') + 10;
		}
		if (i & 1) {
			hexstr[j++] |= c;
		} else {
			hexstr[j] = c << 4;
		}
		i++;
	}
	if (i & 1) {
		fprintf(stderr, "Number of hex digits not even\n");
		return (-1);
	}
	i >>= 1;
	return (i);
}


void
usage(cmdname)
	char		*cmdname;
{
	fprintf(stderr,
	 "Usage: %s [-u <call user data>] <remote_addr> [<file_to_receive>]\n",
		cmdname);
	exit (1);
}


void
main(argc, argv)
	int		argc;
	char		**argv;
{
	char		*cmdname;
	char		*remote_addr;
	int             svc;
	int		fd = 1;		/* Write to stdout by default. */
	char		cud[MAX_CUD_LEN];
	int		cud_len = 0;

	cmdname = argv[0];
	argv++; argc--;
	if (argc < 1) {
		usage(cmdname);
	}
	while (argc) {
		if (argv[0][0] != '-') {
			break;
		}
		switch (argv[0][1]) {
		case 'u':
			if (argc < 2) {
				usage(cmdname);
			}
			cud_len = pack(argv[1], cud);
			if (cud_len < 0) {
				usage(cmdname);
			}
			argv++; argc--;
			break;
		default:
			usage(cmdname);
		}
		argv++; argc--;
	}
	if (argc != 1 && argc != 2) {
		usage(cmdname);
	}
	remote_addr = argv[0];
	if (argc == 2) {
		fd = open(argv[1], O_RDWR | O_CREAT, 0644);
		if (fd < 0) {
			fprintf(stderr, "Can't open '%s' - ", argv[1]);
			perror("");
			exit (1);
		}
	}

	svc = open_stream();
	if (cud_len == 0) {
		cud[0] = 0x12;
		cud[1] = 0x34;
		cud_len = 2;
	}
	start_connection(remote_addr, svc, cud, cud_len);
	data_transfer(svc, fd);
	send_clear_req(svc);

	exit (0);
}
